LIBRARY libnfs
EXPORTS
mount_free_export_list
mount_getexports
mount_getexports_async
nfs_find_local_servers
free_nfs_srvr_list
nfs_access
nfs_access_async
nfs_chmod
nfs_chmod_async
nfs_chown
nfs_chown_async
nfs_close
nfs_close_async
nfs_closedir
nfs_creat
nfs_creat_async
nfs_destroy_context
nfs_fchmod
nfs_fchmod_async
nfs_fchown
nfs_fchown_async
nfs_fstat
nfs_fstat_async
nfs_fsync
nfs_fsync_async
nfs_ftruncate
nfs_ftruncate_async
nfs_get_error
nfs_get_fd
nfs_get_readmax
nfs_get_writemax
nfs_init_context
nfs_link
nfs_link_async
nfs_lseek
nfs_lseek_async
nfs_mkdir
nfs_mkdir_async
nfs_mknod
nfs_mknod_async
nfs_mount
nfs_mount_async
nfs_open
nfs_open_async
nfs_opendir
nfs_opendir_async
nfs_pread
nfs_pread_async
nfs_pwrite
nfs_pwrite_async
nfs_read
nfs_read_async
nfs_readdir
nfs_readlink
nfs_readlink_async
nfs_rename
nfs_rename_async
nfs_rmdir
nfs_rmdir_async
nfs_service
nfs_set_auth
nfs_stat
nfs_stat_async
nfs_statvfs
nfs_statvfs_async
nfs_symlink
nfs_symlink_async
nfs_truncate
nfs_truncate_async
nfs_unlink
nfs_unlink_async
nfs_utime
nfs_utime_async
nfs_utimes
nfs_utimes_async
nfs_which_events
nfs_write
nfs_write_async
